/*
 * function: Peak Normalisation support for Wave files
 *
 * This program is distributed under the GNU General Public License, version 
 * 2.1. A copy of this license is included with this source.
 *
 * Copyright (C) 2002-2004 John Edwards
 * Additional code by Magnus Holmgren and Gian-Carlo Pascutto
 * Linux patch by Marc Brooker
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include "getopt.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "config.h"
#include "i18n.h"
#include "audio.h"
#include "misc.h"
#include "volumax.h"
#include "main.h"
#include "dither.h"

#ifdef _WIN32
#include <windows.h>
#endif

#ifdef ENABLE_RECURSIVE
#include "recurse.h"
#endif

/*This is an ugly way of doing this - but it works for the moment*/
#ifndef MAX_PATH
#define MAX_PATH 512
#endif

int	      write_log = 0;
char          log_file_name[MAX_PATH];
extern double total_files;

/**
 * \brief Allocate a FILE_LIST node.
 *
 * Allocate a FILE_LIST node. filename is set to file while track_peak and
 * track_gain are set to NO_PEAK and NO_GAIN respectively. The rest of the
 * fields are set to zero.
 *
 * \param file  name of file to get a node for.
 * \return  the allocated node or NULL (in which case a message has been
 *          printed).
 */
static FILE_LIST* alloc_node(const char* file)
{
	FILE_LIST* node;

	node = calloc(1, sizeof(*node));

	if (node != NULL) {
		node->filename = strdup(file);

		if (node->filename != NULL) {
			node->track_peak = NO_PEAK;
			node->track_gain = NO_GAIN;
			return node;
		}
		free(node);
	}

	fprintf(stderr, _("Out of memory\n"));
	return NULL;
}


/**
 * \brief Allocate a new FILE_LIST node and add it to the end of a list.
 *
 * Allocate a new FILE_LIST node and add it to the end of a list.
 *
 * \param list  pointer to the list (first node) pointer. If the list
 *              pointer contains NULL, it is set to the new node. Otherwise 
 *              the new node is added the last node in the list.
 * \param file  name of the file to get a node for.
 * \return  0 if successful and -1 if the node couldn't be allocated (in
 *          which case a message has been printed).
 */
int add_to_list(FILE_LIST** list, const char* file)
{
	if (*list == NULL) {
		*list = alloc_node(file);

		if (*list == NULL)
			return -1;
	}
	else {
		FILE_LIST* node = *list;

		while (node->next_file != NULL)
			node = node->next_file;

		node->next_file = alloc_node(file);

		if (node->next_file == NULL)
			return -1;
	}

	return 0;
}


/**
 * Free all nodes in a FILE_LIST list.
 *
 * \param list  pointer to first node in list. Can be NULL, in which case 
 *              nothing is done.
 */
void free_list(FILE_LIST* list)
{
	FILE_LIST* next;

	while (list != NULL) {
		next = list->next_file;
		free((void *) list->filename);
		free(list);
		list = next;
	}
}

/**
 * \brief Processs the file in file_list.
 *
 * Process the files in file_list. If settings->album is set, the files are
 * considered to belong to one album.
 *
 * \param file_list  list of files to process.
 * \param settings   settings and global variables.
 * \return  0 if successful and -1 if an error occured (in which case a
 *          message has been printed).
 */
int process_files(FILE_LIST* file_list, SETTINGS* settings, const char* dir)
{
	FILE_LIST* file;

	settings->album_peak = NO_PEAK;

	if (file_list == NULL)
		return 0;

	settings->first_file = 1;

	/* Analyze the files */
	for (file = file_list; file; file = file->next_file) {
		if (file->filename == '\0')
			continue;

		if (!get_gain(file->filename, &file->track_peak, &file->track_gain, settings)) {
			file->filename = 0;
			continue;
		}
	}

	fprintf(stderr, "\n");
	if(write_log)
		log_error("\n");

	if (settings->audiophile) {
		double scale;
		scale  = settings->album_new_peak / settings->album_peak;
		if (scale < 1.01 && scale > 0.99 && !settings->need_to_process) {
			fprintf(stderr, " No Album Volume adjustment required, exiting.\n");
			if(write_log)
				log_error(" No Album Volume adjustment required, exiting.\n");
			settings->set_album_gain = 1;
		}
		else {
			fprintf(stderr, " Recommended Album Volume adjustment of %5.2lf\n\n", scale);
			if(write_log)
				log_error(" Recommended Album Volume adjustment of %5.2lf\n\n", scale);
		}
	}

	if(settings->apply_gain) {	/* Write radio and audiophile gains. */
		total_files = 0.0;
		for (file = file_list; file; file = file->next_file) {
			if (settings->audiophile && settings->set_album_gain == 1 && !settings->need_to_process)
				break;
			if (settings->radio) {
				double scale;
				scale  = (double) (file->track_gain / file->track_peak);
				if (scale < 1.01 && scale > 0.99 && !settings->need_to_process) {
					fprintf(stderr, " No Title Volume adjustment required for file: %s, skipping.\n", file->filename);
					if(write_log)
						log_error(" No Title Volume adjustment required for file: %s, skipping.\n", file->filename);
				}
				else if (!write_gains(file->filename, &file->track_gain, &file->track_peak, settings)) {
					fprintf(stderr, " Error processing Volume adjustment for file - %s\n", file->filename);
				}
			}
			else if (!write_gains(file->filename, &file->track_gain, &file->track_peak, settings)) {
				fprintf(stderr, " Error processing Volume adjustment for file - %s\n", file->filename);
			}
		}
	}

	fprintf(stderr, "\n Volumax Processing completed normally\n");
	if(write_log)
		log_error("\n Volumax Processing completed normally\n\n");

	return 0;
}


/**
 * Print out a list of options and the command line syntax.
 */
static void usage(void)
{
	fprintf(stdout, _("Volumax v" VOLUMAX_VERSION " Compiled " __DATE__ ".\n"));
	fprintf(stdout, _("Copyright (c) 2002-2008 John Edwards <john.edwards33@ntlworld.com>\n"));
	fprintf(stdout, " Usage: volumax [options] input.wav [...]\n\n");
	fprintf(stdout, " OPTIONS\n");
	fprintf(stdout, "  -h, --help       Prints this help information.\n");
	fprintf(stdout, "  -a, --album      Use Audiophile/Album volume setting, or\n");
	fprintf(stdout, "  -r, --radio      Use Radio/Single Track volume setting(DEFAULT).\n");
	fprintf(stdout, "  -c, --calculate  Calculates and prints proposed volume adjustments, but\n");
	fprintf(stdout, "                   DOES NOT APPLY THEM - This is the DEFAULT.\n");
	fprintf(stdout, "  -y, --apply      Calculates and APPLIES proposed volume adjustments.\n");
	fprintf(stdout, "  -g, --gain X     Apply reduction to maximum non-clipping volume in dB, where\n");
	fprintf(stdout, "             X = any floating point number between 0.0 and -12.0(DEFAULT 0.0)\n");
#ifdef ENABLE_RECURSIVE
	fprintf(stdout, "  -z, --recursive  Search for files recursively, each folder as an album\n");
#endif
	fprintf(stdout, "  -l, --log        Write log file.(Default filename = VLog.txt)\n");
	fprintf(stdout, "  -f, --logfile    Specify log filename. (Assumes -l if present.)\n");
	fprintf(stdout, "  -d, --dither X   Dither output, where X =\n");
	fprintf(stdout, "               0   for       dither OFF (default).\n");
	fprintf(stdout, "               1   for       dither without Noise Shaping.\n");
	fprintf(stdout, "               2   for       dither with Light Noise Shaping.\n");
	fprintf(stdout, "               3   for       dither with Medium Noise Shaping.\n");
	fprintf(stdout, "               4   for       dither with Heavy Noise Shaping.\n");
	fprintf(stdout, "  -o, --stdout     Write output file to stdout.\n");
	fprintf(stdout, " FORMAT OPTIONS (One option ONLY may be used)\n");
	fprintf(stdout, "  -b, --bits X     Set output sample format, where X =\n");
	fprintf(stdout, "             1     for        8 bit unsigned PCM data.\n");
	fprintf(stdout, "             2     for       16 bit signed PCM data.\n");
	fprintf(stdout, "             3     for       24 bit signed PCM data.\n");
	fprintf(stdout, "             4     for       32 bit signed PCM data.\n");
	fprintf(stdout, "             5     for       32 bit floats.\n");
	fprintf(stdout, "             6     for       16 bit 'aiff' format.\n");
	fprintf(stdout, "         NOTE:     By default, the output file will be of the same bitwidth\n");
	fprintf(stdout, "                   and type as the input file.\n");
	fprintf(stdout, " INPUT FILES\n");
	fprintf(stdout, "  Volumax input files may be 8, 16, 24 or 32 bit integer, or floating point\n"); 
	fprintf(stdout, "  wave files with 1 or 2 channels and a sample rate of 48000Hz, 44100Hz,\n");
	fprintf(stdout, "  32000Hz, 24000Hz, 22050Hz, 16000Hz, 12000Hz, 11025Hz or 8000Hz.\n");
	fprintf(stdout, "  16 bit integer 'aiff' files are also supported.\n");
	fprintf(stdout, "  Wildcards (?, *) can be used in the filename, or '-' for stdin.\n");

    return;
}


const static struct option long_options[] = {
	{"help",	0, NULL, 'h'},
	{"album",	0, NULL, 'a'},
	{"radio",	0, NULL, 'r'},
	{"calculate",	0, NULL, 'c'},
	{"apply",	0, NULL, 'y'},
	{"stdout",	0, NULL, 'o'},
#ifdef ENABLE_RECURSIVE
	{"recursive",   0, NULL, 'z'},
#endif
	{"gain",	1, NULL, 'g'},
	{"log",		0, NULL, 'l'},
	{"logfile",	1, NULL, 'f'},
	{"dither",	1, NULL, 'd'},
	{"bits",	1, NULL, 'b'},
	{NULL,		0, NULL , 0}
};


#ifdef ENABLE_RECURSIVE
#define ARG_STRING "harcyozg:lf:d:b:"
#else
#define ARG_STRING "harcyog:lf:d:b:"
#endif


int main(int argc, char** argv)
{
	SETTINGS settings;
	int      option_index = 1,
	         ret,
	         i,
	         bits;

 	char     CmdDir[MAX_PATH];
 	char     *p;

	memset(&settings, 0, sizeof(settings));
	settings.first_file = 1;
	settings.radio = 1;
	settings.outbitwidth = 16;
	settings.format = WAV_NO_FMT;

#ifdef _WIN32
	/* Is this good enough? Or do we need to consider multi-byte codepages as 
	* well? 
	*/
	SetConsoleOutputCP(GetACP());

	GetModuleFileName(NULL, CmdDir, MAX_PATH);
	p = strrchr(CmdDir, '\\') + 1;
	p[0] = '\0';
#endif

	while ((ret = getopt_long(argc, argv, ARG_STRING, long_options, &option_index)) != -1) {
		switch(ret) {
			case 0:
				fprintf(stderr, "Internal error parsing command line options\n");
				exit(1);
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'a':
				settings.audiophile = 1;
				break;
			case 'r':
				settings.radio = 1;
				break;
			case 'c':
				settings.apply_gain = 0;
				break;
			case 'y':
				settings.apply_gain = 1;
				break;
			case 'o':
				settings.std_out = 1;
				break;
#ifdef ENABLE_RECURSIVE
			case 'z':
				settings.recursive = 1;
				break;
#endif
			case 'l':
				write_log = 1;
#ifdef _WIN32
				strcpy(log_file_name, CmdDir);
				strcat(log_file_name, LOG_NAME);
#else
				strcpy(log_file_name, LOG_NAME);
#endif
				break;
			case 'f':
				write_log = 1;
				strcpy(log_file_name, optarg);
				break;
			case 'd':
				settings.need_to_process = 1;
				settings.dithering = 1;
   				if(sscanf(optarg, "%d", &settings.shapingtype) != 1) {
	    				fprintf(stderr, "Warning: dither type %s not recognised, using default\n", optarg);
		    			break;
				}
				if (settings.shapingtype == 0)
					settings.dithering = 0;
				else if (settings.shapingtype == 1)
					settings.shapingtype = 0;
				else if (settings.shapingtype == 2)
					settings.shapingtype = 1;
				else if (settings.shapingtype == 3)
					settings.shapingtype = 2;
				else if (settings.shapingtype == 4)
					settings.shapingtype = 3;
				break;
			case 'g':
   				if(sscanf(optarg, "%lf", &settings.man_gain) != 1) {
	    				fprintf(stderr, "Warning: manual gain %s not recognised, ignoring\n", optarg);
					break;
				}
				if(settings.man_gain < -12.0) {
	    				fprintf(stderr, "Warning: manual gain %s is out of range, "
					                 "applying gain of -12.0dB\n", optarg);
					settings.man_gain = -12.0;
				}
				else if(settings.man_gain > 0.0) {
	    				fprintf(stderr, "Warning: manual gain %s is out of range, "
					                 "ignoring\n", optarg);
					settings.man_gain = 0.0;
				}
				break;
			case 'b':
				settings.need_to_process = 1;
   				if(sscanf(optarg, "%d", &bits) != 1) {
	    				fprintf(stderr, "Warning: output format %s not recognised, using default\n", optarg);
		    			break;
				}
				if (bits == 1) {
					settings.outbitwidth = 8;
					settings.format = WAV_FMT_8BIT;
		    			break;
				}
				else if (bits == 2) {
					settings.outbitwidth = 16;
					settings.format = WAV_FMT_16BIT;
		    			break;
				}
				else if (bits == 3) {
					settings.outbitwidth = 24;
					settings.format = WAV_FMT_24BIT;
		    			break;
				}
				else if (bits == 4) {
					settings.outbitwidth = 32;
					settings.format = WAV_FMT_32BIT;
		    			break;
				}
				else if (bits == 5) {
					settings.outbitwidth = 32;
					settings.format = WAV_FMT_FLOAT;
		    			break;
				}
				else if (bits == 6) {
					settings.outbitwidth = 16;
					settings.format = WAV_FMT_AIFF;
		    			break;
				}
				else {
	    				fprintf(stderr, "Warning: output format %s not recognised, using default\n", optarg);
					break;
				}
				break;
		}
	}


	if (optind >= argc) {
		fprintf(stderr, _("No files specified.\n"));
		usage();
		return EXIT_SUCCESS;
	}

	if (write_log) {
		log_error("Command line:\n\t");
		for(i = 0; i < argc; i++)
			log_error("%s ",  argv[i]);
		log_error("\n");
	}

	if (!strcmp(argv[optind], "-")) {
		double track_peak, track_gain;
		if (!get_gain("-", &track_peak, &track_gain, &settings))
			return -1;
	}
	else {
		for (i = optind; i < argc; ++i) {
#ifdef ENABLE_RECURSIVE
			if (process_argument(argv[i], &settings) < 0) {
				free_list(settings.file_list);
				return EXIT_FAILURE;
			}
#else
			if (add_to_list(&settings.file_list, argv[i]) < 0)
				return EXIT_FAILURE;
#endif
		}

		/* Process files (perhaps remaining) in list */
		ret = process_files(settings.file_list, &settings, ".");
		free_list(settings.file_list);
		settings.file_list = NULL;
	}

	return (ret < 0) ? EXIT_FAILURE : EXIT_SUCCESS;
}
