/*
 * function: Peak volume normalisation for Wave files
 *
 * This program is distributed under the GNU General Public License, version 
 * 2.1. A copy of this license is included with this source.
 *
 * Copyright (C) 2002-2015 John Edwards
 * Linux patch by Marc Brooker
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#ifdef _WIN32
#include <io.h>
#else
# ifndef __MACOSX__
#  include <sys/io.h>
# endif
#endif

#include <fcntl.h>

#ifndef __MACOSX__
#include <malloc.h>
#endif

#include "i18n.h"
//#include "config.h"
#include "getopt.h"
#include "misc.h"
#include "audio.h"
#include "dither.h"
#include "main.h"
#include "volumax.h"

#ifdef _WIN32
#include <windows.h>
#endif

#ifdef ENABLE_RECURSIVE
#include "recurse.h"
#endif

/*Gcc uses LL as a suffix for long long int (64 bit) types - Marc Brooker 8/4/2004*/
#ifdef __GNUC__
#define ROUND64(x)   ( doubletmp = (x) + Dither.Add + (Int64_t)0x001FFFFD80000000LL, *(Int64_t*)(&doubletmp) - (Int64_t)0x433FFFFD80000000LL )
#else
#define ROUND64(x)   ( doubletmp = (x) + Dither.Add + (Int64_t)0x001FFFFD80000000L, *(Int64_t*)(&doubletmp) - (Int64_t)0x433FFFFD80000000L )
#endif

extern int          write_log;
dither_t            Dither;
double              doubletmp;
double              total_samples;
double              total_files;

/* Replaced with a double based function for consistency 2005-11-17
static float FABS(float x)
{
	unsigned int *ix=(unsigned int *)&x;
	*ix&=0x7fffffffUL;
	return(x);
}
*/
static double DABS(double x)
{
	Uint64_t *ix=(Uint64_t *)&x;
#ifdef __GNUC__
	*ix&=0x7fffffffffffffffULL;
#else
	*ix&=0x7fffffffffffffff;
#endif
	return(x);
}

/* Dither output */
Int64_t dither_output(int dithering, int shapingtype, int i, double Sum, int k, int format)
{
	double Sum2;
	Int64_t val;
	if(dithering) {
		if(!shapingtype) {
			double  tmp = Random_Equi ( Dither.Dither );
			Sum2 = tmp - Dither.LastRandomNumber [k];
			Dither.LastRandomNumber [k] = (int)tmp;
			Sum2 = Sum += Sum2;
			val = ROUND64 (Sum2)  &  Dither.Mask;
		}
		else {
			Sum2  = Random_Triangular ( Dither.Dither ) - scalar16 ( Dither.DitherHistory[k], Dither.FilterCoeff + i );
			Sum  += Dither.DitherHistory [k] [(-1-i)&15] = (float)Sum2;
			Sum2  = Sum + scalar16 ( Dither.ErrorHistory [k], Dither.FilterCoeff + i );
			val = ROUND64 (Sum2)  &  Dither.Mask;
			Dither.ErrorHistory [k] [(-1-i)&15] = (float)(Sum - val);
		}
	}
	else
		val = (Int64_t)(ROUND64 (Sum));

	if (format == WAV_FMT_8BIT)
		val = val >> 24;
	else if (format == WAV_FMT_16BIT || format == WAV_FMT_AIFF)
		val = val >> 16;
	else if (format == WAV_FMT_24BIT)
		val = val >> 8;

	return (val);
}

/* Get the gain and peak value for a file. Runs in audiophile mode if 
 * audiophile is true.
 *
 * If an error occured, 0 is returned (a message has been printed).
 */

int get_gain(const char *filename, double *track_peak, double *track_gain, SETTINGS *settings)
{
	volumax_opt *vm_opts = malloc(sizeof(volumax_opt));
	FILE         *infile;
	int          result = 0;
	double       new_peak,
	             factor_clip;
	int          i;
	input_format *format;

	if(!strcmp(filename, "-")) {
		infile = stdin;
		settings->apply_gain = 0;
#ifdef _WIN32
		_setmode( _fileno(stdin), _O_BINARY );
#endif
	}
	else
		infile = fopen(filename, "rb");

	if (infile == NULL) {
		fprintf (stderr, " Not able to open input file %s.\n", filename) ;
		goto exit;
	}
	vm_opts->apply_gain = 0;

	/*
	 * Now, we need to select an input audio format
	 */

	format = open_audio_file(infile, vm_opts);
	if (!format) {
		/* error reported by reader */
		fprintf (stderr, " Unrecognized file format for %s.\n", filename) ;
		goto exit;
	}

	if ((vm_opts->channels != 1) && (vm_opts->channels != 2)) {
		fprintf(stderr, " Unsupported number of channels.\n");
		goto exit;
	}

	if (settings->first_file) {
		total_samples = (double)vm_opts->total_samples_per_channel;
		fprintf(stderr, "\n Analyzing...\n\n");
		fprintf(stderr, "  Peak  | Scale | New Peak | Track\n");
		fprintf(stderr, " -------------------------------------\n");
		if(write_log) {
			log_error("\n Analyzing...\n\n");
			log_error("  Peak  | Scale | New Peak | Track\n");
			log_error(" -------------------------------------\n");
		}
		settings->first_file = 0;
	}
	else
		total_samples += (double)vm_opts->total_samples_per_channel;

	{
		long samples_read;
		double **buffer = malloc(sizeof(double *) * vm_opts->channels);

		for (i = 0; i < vm_opts->channels; i++)
			buffer[i] = malloc(BUFFER_LEN * sizeof(double));

		while (1) {

			samples_read = vm_opts->read_samples(vm_opts->readdata, buffer, BUFFER_LEN, 0, 0);

			if (samples_read == 0) {
				break;
			} 
			else {
				if (samples_read < 0) {
					/* Error in the stream. Not a problem, just reporting it in case 
					 * we (the app) cares. In this case, we don't
					 */
				} 
				else {
					int i;
					int j;

					for (i = 0; i < vm_opts->channels; i++) {
						for (j = 0; j < samples_read; j++) {
							buffer[i][j] *= 0x7fff;
							if (DABS(buffer[i][j]) > *track_peak)
								*track_peak = DABS(buffer[i][j]);
						}
					}
				}
			}
		}

		for (i = 0; i < vm_opts->channels; i++) {
			if (buffer[i]) free(buffer[i]);
		}
		if (buffer) free(buffer);
	}

	/*
	 * calculate factors for max non-Clipping level
	 */

	*track_gain = (double) (pow(10., settings->man_gain * 0.05) * 32767.);

	factor_clip  = (double) (*track_gain / *track_peak);

	new_peak = (double) *track_peak * factor_clip;

	fprintf(stderr, " %6.0f | %5.2f | %8.0f | %s\n",
		*track_peak, factor_clip, new_peak, filename);
	if(write_log) {
		log_error(" %6.0f | %5.2f | %8.0f | %s\n",
			*track_peak, factor_clip, new_peak, filename);
	}
			
	settings->album_new_peak = settings->album_new_peak < new_peak ? new_peak : settings->album_new_peak;
	settings->album_peak = settings->album_peak < *track_peak ? *track_peak : settings->album_peak;

	result = 1;

exit:
	if (result)
		format->close_func(vm_opts->readdata);
	if (vm_opts)
		free(vm_opts);
	if (infile)
		fclose(infile);
	return result;
}


/* Use the ReplayGain calculations to adjust the gain on the wave file.
 * If audiophile_gain is selected, that value is used, otherwise the
 * radio_gain value is used.
 */
int write_gains(const char *filename, double *radio_gain, double *TitlePeak, SETTINGS *settings)
{
	volumax_opt *vm_opts = malloc(sizeof(volumax_opt));
	FILE         *infile;
	audio_file   *aufile;
	int          readcount,
	             result = 0,
	             delete_temp = 0,
	             i;
	double       scale;
	double       total_read = 0.;
	double       wrap_prev_pos;
	double       wrap_prev_neg;
	void         *sample_buffer;
	input_format *format;
	char         *temp_file;
	
	temp_file = NULL;

	memset(vm_opts, 0, sizeof(volumax_opt));

	infile = fopen(filename, "rb");

	if (infile == NULL) {
		fprintf (stderr, " Not able to open input file %s.\n", filename) ;
		goto exit;
	}
	vm_opts->apply_gain = 1;

	/*
	 * Now, we need to select an input audio format
	 */

	format = open_audio_file(infile, vm_opts);
	if (!format) {
		format->close_func(vm_opts->readdata);
		if (vm_opts)
			free(vm_opts);
		fclose(infile);
		/* error reported by reader */
		fprintf (stderr, " Unrecognized file format for %s.\n", filename) ;
	}
	else {
		double **pcm = malloc(sizeof(double *) * vm_opts->channels);

		for (i = 0; i < vm_opts->channels; i++)
			pcm[i] = malloc(BUFFER_LEN * sizeof(double));

		switch(settings->format) {
			case WAV_NO_FMT:
				if (vm_opts->format == WAV_FMT_AIFF || vm_opts->format == WAV_FMT_AIFC8
								   || vm_opts->format == WAV_FMT_AIFC16) {
					vm_opts->format = WAV_FMT_AIFF;
					wrap_prev_pos = 0x7fff;
					wrap_prev_neg = -0x8000;
				}
				else if (vm_opts->format == WAV_FMT_8BIT) {
					wrap_prev_pos = 0x7f;
					wrap_prev_neg = -0x80;
				}
				else if (vm_opts->format == WAV_FMT_16BIT) {
					wrap_prev_pos = 0x7fff;
					wrap_prev_neg = -0x8000;
				}
				else if (vm_opts->format == WAV_FMT_24BIT) {
					wrap_prev_pos = 0x7fffff;
					wrap_prev_neg = -0x800000;
				}
				else if (vm_opts->format == WAV_FMT_32BIT) {
					wrap_prev_pos = 0x7fffffff;
					wrap_prev_neg = -0x7fffffff;
				}
				else if (vm_opts->format == WAV_FMT_FLOAT) {
					wrap_prev_pos = 1 - (1 / 0x80000000);
					wrap_prev_neg = -1.;
				}
				break;
			case WAV_FMT_8BIT:
				vm_opts->format = WAV_FMT_8BIT;
				vm_opts->samplesize = 8;
				wrap_prev_pos = 0x7f;
				wrap_prev_neg = -0x80;
				break;
			case WAV_FMT_AIFF:
				vm_opts->format = WAV_FMT_AIFF;
				vm_opts->samplesize = 16;
				vm_opts->endianness = BIG;
				wrap_prev_pos = 0x7fff;
				wrap_prev_neg = -0x8000;
				break;
			case WAV_FMT_16BIT:
				vm_opts->format = WAV_FMT_16BIT;
				vm_opts->samplesize = 16;
				vm_opts->endianness = LITTLE;
				wrap_prev_pos = 0x7fff;
				wrap_prev_neg = -0x8000;
				break;
			case WAV_FMT_24BIT:
				vm_opts->format = WAV_FMT_24BIT;
				vm_opts->samplesize = 24;
				vm_opts->endianness = LITTLE;
				wrap_prev_pos = 0x7fffff;
				wrap_prev_neg = -0x800000;
				break;
			case WAV_FMT_32BIT:
				vm_opts->format = WAV_FMT_32BIT;
				vm_opts->samplesize = 32;
				vm_opts->endianness = LITTLE;
				wrap_prev_pos = 0x7fffffff;
				wrap_prev_neg = -0x7fffffff;
				break;
			case WAV_FMT_FLOAT:
				vm_opts->format = WAV_FMT_FLOAT;
				vm_opts->samplesize = 32;
				vm_opts->endianness = LITTLE;
				wrap_prev_pos = 1 - (1 / 0x80000000);
				wrap_prev_neg = -1.;
				break;
		}

		vm_opts->std_out = settings->std_out;
		
		temp_file = malloc( (strlen(filename) + strlen(TEMP_NAME) + 16) * sizeof(char) );
		sprintf(temp_file, "%s.%s", filename, TEMP_NAME);

		aufile = open_output_audio_file(temp_file, vm_opts);

		if (aufile == NULL) {
			fprintf (stderr, " Not able to open output file %s.\n", temp_file);
			fclose(infile);
			goto exit;
		}

		Init_Dither (vm_opts->samplesize, settings->shapingtype);

		if(settings->audiophile)
			scale  = (double) (*radio_gain / settings->album_peak);
		else
			scale  = (double) (*radio_gain / *TitlePeak);

		fprintf(stderr, " Applying Volume Adjustment of %5.2lf to file: %s\n", scale, filename);
		if (write_log) {
			log_error(" Applying Volume Adjustment of %5.2lf to file: %s\n", scale, filename);
		}

		while (1) {

			readcount = vm_opts->read_samples(vm_opts->readdata, pcm, BUFFER_LEN, 0, 0);

			total_read += ((double)readcount / vm_opts->rate);
			total_files += ((double)readcount / vm_opts->rate);
			if( (long)total_files % 4 == 0) {
				fprintf(stderr, "This file %3.0lf%% done\tAll files %3.0lf%% done\r", 
					total_read / (vm_opts->total_samples_per_channel / vm_opts->rate) * 100,
					total_files / (total_samples / vm_opts->rate) * 100);
			}

			if (readcount == 0) {
				break;
			} 
			else if (readcount < 0) {
				/* Error in the stream. Not a problem, just reporting it in case 
				 * we (the app) cares. In this case, we don't
				 */
			} 
			else {
				int   convsize = BUFFER_LEN;
				int   j,
				      i = 0,
				      k;
				int   bout = (readcount < convsize ? readcount : convsize);

				/* scale doubles to 8, 16, 24 or 32 bit signed ints 
				 * (host order) (unless float output)
				 * and apply ReplayGain scaling, etc. 
				 */
				sample_buffer = malloc(sizeof(double) * vm_opts->channels * bout);
				for(k = 0; k < vm_opts->channels; k++) {
					for(j = 0; j < bout; j++, i++) {
						Int64_t val;
						double Sum;

						pcm[k][j] *= scale;
						if (vm_opts->format != WAV_FMT_FLOAT) {
							Sum = pcm[k][j]*2147483647.f;
							if (i > 31)
								i = 0;
							val = dither_output(settings->dithering, settings->shapingtype, i,
									    Sum, k, vm_opts->format);
							if (val > (Int64_t)wrap_prev_pos)
								val = (Int64_t)wrap_prev_pos;
							else if (val < (Int64_t)wrap_prev_neg)
								val = (Int64_t)wrap_prev_neg;
							pcm[k][j] = (double)val;
						}
						else {
							if (pcm[k][j] > wrap_prev_pos)
								pcm[k][j] = wrap_prev_pos;
							else if (pcm[k][j] < wrap_prev_neg)
								pcm[k][j] = wrap_prev_neg;
						}
					}
				}
				sample_buffer = output_to_PCM(pcm, sample_buffer, vm_opts->channels,
						bout, vm_opts->format);
				/* write to file */
				write_audio_file(aufile, sample_buffer, bout * vm_opts->channels);

				if (sample_buffer) free(sample_buffer);
			}
		}
		for (i = 0; i < vm_opts->channels; i++)
			if (pcm[i]) free(pcm[i]);
		if (pcm) free(pcm);
		format->close_func(vm_opts->readdata);
		close_audio_file(infile, aufile, vm_opts);
		if (infile) fclose(infile);

		if (!settings->std_out) {
			if (remove(filename) != 0) {
				fprintf(stderr, " Error deleting old file '%s'\n", filename);
				goto exit;
			}
    
			if (rename(temp_file, filename) != 0) {
				fprintf(stderr, " Error renaming '%s' to '%s' (uh-oh)\n", temp_file, filename);
				goto exit;
			}
		}
    
		result = 1;
	}
exit:
	if(temp_file != NULL) free(temp_file);
	return result;
}


